@@ -3,6 +3,7 @@ import logging
 from collections import namedtuple
 from datetime import datetime
 
+from django.conf import settings
 from django.db import transaction
 from django.db.models import Max, Q
 from django.utils import timezone
@@ -77,12 +78,40 @@ def pre_migrate_content_type(content_model, mutable_type, lazy_type, premigrate_
         content_model: Models for content which is being migrated.
         mutable_type: Boolean that indicates whether the content type is mutable.
     """
-    batch_size = 100
+    def delete_removed_pulp2_content(content_model):
+        """
+        Delete Pulp2Content records for content which is no longer present in Pulp2.
+
+        This is to avoid situations and extra work when not all content migrated during the first
+        migration run, then orphan clean up is run in Pulp 2, and then migration is run again.
+
+        Args:
+            content_model: Pulp 2 content model
+
+        """
+        content_type = content_model.pulp2.TYPE_ID
+        mongo_content_qs = content_model.pulp2.objects().only('id')
+        mongo_content_ids = {c['_id'] for c in mongo_content_qs.as_pymongo().no_cache()}
+        premigrated_content_ids = set(
+            Pulp2Content.objects.filter(
+                pulp2_content_type_id=content_type
+            ).only('pulp2_id').values_list('pulp2_id', flat=True)
+        )
+        content_ids_to_delete = premigrated_content_ids - mongo_content_ids
+        if content_ids_to_delete:
+            Pulp2Content.objects.filter(
+                pulp2_content_type_id=content_type,
+                pulp2_id__in=content_ids_to_delete
+            ).delete()
+
+    batch_size = settings.CONTENT_PREMIGRATION_BATCH_SIZE or DEFAULT_BATCH_SIZE
     pulp2content = []
     pulp2mutatedcontent = []
     content_type = content_model.pulp2.TYPE_ID
     set_pulp2_repo = content_model.pulp_2to3_detail.set_pulp2_repo
 
+    delete_removed_pulp2_content(content_model)
+
     # the latest timestamp we have in the migration tool Pulp2Content table for this content type
     content_qs = Pulp2Content.objects.filter(pulp2_content_type_id=content_type)
     last_updated = content_qs.aggregate(Max('pulp2_last_updated'))['pulp2_last_updated__max'] or 0
@@ -105,13 +134,13 @@ def pre_migrate_content_type(content_model, mutable_type, lazy_type, premigrate_
         total=total_content))
 
     pulp2content_pb = ProgressReport(
-        message='Pre-migrating Pulp 2 {} content (general info)'.format(content_type.upper()),
+        message='Pre-migrating Pulp 2 {} content (general info)'.format(content_type),
         code='premigrating.content.general',
         total=total_content,
         state=TASK_STATES.RUNNING)
     pulp2content_pb.save()
     pulp2detail_pb = ProgressReport(
-        message='Pre-migrating Pulp 2 {} content (detail info)'.format(content_type.upper()),
+        message='Pre-migrating Pulp 2 {} content (detail info)'.format(content_type),
         code='premigrating.content.detail',
         total=total_content,
         state=TASK_STATES.RUNNING)
@@ -157,7 +186,10 @@ def pre_migrate_content_type(content_model, mutable_type, lazy_type, premigrate_
                 pulp2detail_pb.total -= 1
                 continue
 
-        downloaded = record.downloaded if hasattr(record, 'downloaded') else False
+        # very old pulp2 content will not have downloaded field set (prior to lazy sync)
+        downloaded = hasattr(record, 'downloaded') and (
+            record.downloaded or record.downloaded is None
+        )
 
         if set_pulp2_repo:
             # This content requires to set pulp 2 repo. E.g. for errata, because 1 pulp2
@@ -242,6 +274,79 @@ def pre_migrate_content_type(content_model, mutable_type, lazy_type, premigrate_
             pulp2content.clear()
             existing_count = 0
 
+    # If it's a per-repo content type and it's a migration re-run, we need to make sure that the
+    # existing content hasn't been associated with a new repo since our last migration,
+    # and if so, we need to go back and create a Pulp2Content for these new relations.
+    # E.g. errata copied from one repo to another in Pulp 2, in such cases _last_updated is
+    # unchanged.
+    if set_pulp2_repo and last_updated:
+        # last_updated is a unix timestamp, we need to convert it to use in our Django query.
+        last_updated = datetime.utcfromtimestamp(last_updated)
+
+        # Query all new relations for that content since the last run
+        content_relations = Pulp2RepoContent.objects.filter(
+            pulp2_content_type_id=content_type,
+            pulp2_repository__not_in_plan=False,
+            pulp2_created__gte=last_updated
+        ).select_related(
+            'pulp2_repository'
+        ).only(
+            'pulp2_repository', 'pulp2_created',
+        ).order_by('pulp2_created')
+
+        mongo_content_qs = content_model.pulp2.objects(
+            id__in=content_relations.values_list('pulp2_unit_id', flat=True))
+        pulp2_content_by_id = {
+            record.id: record for record in mongo_content_qs.only(*mongo_fields).no_cache()
+        }
+
+        for relation in content_relations:
+            record = pulp2_content_by_id[relation.pulp2_unit_id]
+            # very old pulp2 content will not have downloaded field set (prior to lazy sync)
+            downloaded = hasattr(record, 'downloaded') and (
+                record.downloaded or record.downloaded is None
+            )
+            specific_content_q = Q(
+                pulp2_content_type_id=record._content_type_id,
+                pulp2_id=record.id,
+                pulp2_repo=relation.pulp2_repository,
+                pulp2_subid='',
+            )
+
+            # Ensure that no existing pulp2content slipped into bulk_create.
+            # Otherwise, we'll have a problem with later bulk_create for detail models.
+            if Pulp2Content.objects.filter(specific_content_q).exists():
+                continue
+
+            item = Pulp2Content(
+                pulp2_id=record.id,
+                pulp2_content_type_id=record._content_type_id,
+                # Set `pulp2_last_updated` to the date of when a content unit got copied.
+                # (We can't set it to anything higher, in case pre-migration crashes and we would
+                # need to pick it up correctly on the next re-run.)
+                # When erratum is copied in pulp 2, it doesn't change its _last_updated timestamp.
+                # It means that Katello has no way to identify that the erratum has been copied
+                # since the last migration run, without reimporting all errata, which is expensive.
+                pulp2_last_updated=int(relation.pulp2_created.timestamp()),
+                pulp2_storage_path=record._storage_path,
+                downloaded=downloaded,
+                pulp2_repo=relation.pulp2_repository
+            )
+            _logger.debug(
+                'Add content item to the list to migrate: {item}'.format(item=item))
+            pulp2content.append(item)
+            pulp2content_pb.total += 1
+            pulp2detail_pb.total += 1
+
+        pulp2content_batch = Pulp2Content.objects.bulk_create(pulp2content)
+        pulp2content_pb.done += len(pulp2content_batch)
+        pulp2content_pb.save()
+
+        content_model.pulp_2to3_detail.pre_migrate_content_detail(pulp2content_batch)
+
+        pulp2detail_pb.done += len(pulp2content_batch)
+        pulp2detail_pb.save()
+
     pulp2content_pb.save()
     pulp2detail_pb.save()
 
@@ -617,8 +722,9 @@ def handle_outdated_resources(plan):
     Args:
         plan(MigrationPlan): A Migration Plan
     """
-    RepoSetup.mark_changed_relations()
-    RepoSetup.finalize()
+    plugins = plan.get_plugins()
+    RepoSetup.mark_changed_relations(plugins)
+    RepoSetup.finalize(plugins)
 
     for plugin_plan in plan.get_plugin_plans():
         inplan_repos = plugin_plan.get_repositories()
